import os
import time
import concurrent.futures

# Import function from module
from .ip_addr_validator import ip_addr_validator
from .file_validator import device_info_list_validator
from .device_info_validator import device_info_validator
from .program_supplementals import enter_key_only, exception_translator, program_cancellation

# Import function from 3rd party module
from pythonping import ping

def connection_ping(ip):
    try:
        # Strip newline at the end of IP address
        ip_addr = ip.rstrip("\n")
        # Ping every 3 seconds for 10 times
        replies = ping(ip_addr, count = 1, interval = 1)
        # No reply is not accepted
        # 3 = SuccessOn.All = All pings are replies
        if replies.success(3) == False:
            # Progress dot
            print(".", end = "", flush = True)
            # Pass values to threading result
            return "FAIL3", ip_addr
        # Less than 6 replies are not accepted
        # 2 = SuccessOn.Most = The total reply is half of the total ping + 1
        elif replies.success(2) == False:
            # Progress dot
            print(".", end = "", flush = True)
            # Pass values to threading result
            return "FAIL2", ip_addr
        # All pings are replies or more than 5 replies are accepted
        else:
            # Progress dot
            print(".", end = "", flush = True)
            # Pass values to threading result
            return "PASS", ip_addr
    except:
        # Progress dot
        print(".", end="", flush=True)
        # Execute exception_translator
        exception_explained = exception_translator()
        # Pass values to threading result
        return "FAIL4", exception_explained

def connection_validator(flag, ip_addr_list):
    # Progress dot with threading capability
    print("\nConcurrently determining per", min(32, os.cpu_count() + 4), "IP addresses. Please wait", end = "", flush = True)
    # Ping-threading stopwatch start
    threading_start = time.time()
    # Ping-threading process
    with concurrent.futures.ThreadPoolExecutor() as executor:
        # Execute threading process
        try:
            # Initial variables
            threads = []
            thread_results = []
            # Loop for every IP address in the list
            for ip_addr in ip_addr_list:
                # Execute ping test for every IP address in the list concurrently
                threads.append(executor.submit(connection_ping, ip = ip_addr))
            # Loop for every result from ping-threading process
            for thread in concurrent.futures.as_completed(threads):
                thread_results.append(thread.result())
        # Cancel the remaining threads by keyboard (e.g. CTRL+C)
        except KeyboardInterrupt:
            executor.shutdown(wait = True, cancel_futures = True)
            # Execute program_cancellation
            program_cancellation()
    # Ping-threading stopwatch end
    print("\n\nPing-threading processed for:", "%.2f" % (time.time() - threading_start), "secs")
    # Loop for every result from ping-threading process results
    if bool([True for result in thread_results if "PASS" not in result[0]]) == True:
        # Expect the raised exception flag
        if "FAIL4" in thread_results[0][0]:
            try:
                # Print the raised exception error messages values
                print("\nFAIL: " + thread_results[0][1][0] + "\n\n" + thread_results[0][1][1])
                # Execute user confirmation to retry connection test
                print("\nPress \'Enter\' to retry connection test or \'CTRL+C\' to end the program", end="", flush=True)
                # Expect the user to press Enter key
                enter_key_only()
                # Repeat execute connection_validator with a flag
                connection_validator("FAIL", ip_addr_list)
            except KeyboardInterrupt:
                # Execute program_cancellation
                print("")
                program_cancellation()
        else:
            # Loop for every result from ping-threading process results. All pings are replies or more than 5 replies are accepted
            print("PASS: Reachable IP address(es): " + str([result[1] for result in thread_results if "PASS" in result[0]])[1:-1])
            # Loop for every result from ping-threading process results. Less than 6 replies are not accepted
            print("FAIL: Flapping IP address(es): " + str([result[1] for result in thread_results if "FAIL2" in result[0]])[1:-1])
            # Loop for every result from ping-threading process results. No reply is not accepted
            print("FAIL: Timed IP address(es): " + str([result[1] for result in thread_results if "FAIL3" in result[0]])[1:-1])
            print("FAIL: Please remove flapping/timed IP address(es) in the list and try again!")
            # Repeat execute device_info_list_validator then device_info_validator then ip_addr_validator with a fail flag and then connection_validator with a flag
            return connection_validator("FAIL", ip_addr_validator("FAIL", device_info_validator(device_info_list_validator())[0]))
    # All pings are replies or more than 5 replies are accepted
    else:
        # IP address list with correction flag
        if flag == "FAIL":
            print("PASS: IP address(es) in the list is reachable")
            # Pass the value
            return ip_addr_list
        # IP address list with no correction flag
        elif flag == "PASS":
            print("PASS: IP address(es) in the list is reachable")


def malas():
    # Import function from module
    from .program_supplementals import program_requirements

    # Execute program_requirements
    program_requirements()

    # Import function from module
    from .connection_ssh import connection_futures
    from .connection_validator import connection_validator
    from .device_info_validator import device_info_validator
    from .configuration_type_validator import configuration_type_validator
    from .ip_addr_validator import ip_addr_validator, ip_addr_list_validator
    from .file_validator import device_info_list_validator, configuration_file_validator
    from .program_supplementals import exit_and_cleanup, enter_key_confirmation, powered_by, program_cancellation

    print("\n \  Malas (Multi-vendor automation leverage at slouching)")
    print("  \  Version 1.0 build 210926.0001 by Ade Destrianto")
    print("   \__________________________________________________________________\n")
    print("Please follow the instructions below. Thank you!\n")
    print("1. The device information list file\'s format is a set of three\n   that contains IP address, username (has privilege), and password separated by double colons \'::\'\n   per line in row formats, for example (Ignore the bracket):\n")
    print("   #==== Device Information List.txt ====#\n   #10.0.0.2::netadmin::password         #\n   #10.0.0.3::netadmin::password         #\n   #=====================================#\n")
    print("2. The configuration file\'s format is one command per line in row formats, for example (Ignore the bracket):\n")
    print("   #========= Configuration.txt =========#\n   #interface GigabitEthernet0/11        #\n   #description Guest_Networks           #\n   #IP address 192.168.1.1 255.255.255.0 #\n   #no shutdown                          #\n   #exit                                 #\n   #do write                             #\n   #=====================================#\n")
    print("3. What is \'Malas\'?")
    print("   Malas is a word from Bahasa Indonesia. The \'Malas\' word meaning is \'Lazy\' (Too inclined to avoid hard work.)\n")

    # Execute user confirmation to continue, then execute configuration_type_validator
    try:
        print("4. Shall we continue?\n   ", end="")
        # Execute enter_key_confirmation
        enter_key_confirmation()
        print("")
        configuration = configuration_type_validator()
        device = configuration[0]
        command = configuration[1]
        delay = configuration[2]
    # Stop process by keyboard (e.g. CTRL+C)
    except KeyboardInterrupt:
        # Execute program_cancellation
        print("")
        program_cancellation()

    # Execute file_validator
    try:
        print(" \  File validator")
        print("  \___________________________________________________________________")
        device_info_list = device_info_list_validator()
        command_list = configuration_file_validator(command)
    # Stop process by keyboard (e.g. CTRL+C)
    except KeyboardInterrupt:
        # Execute program_cancellation
        program_cancellation()

    # Execute device_info_validator
    try:
        print("\n \  Device information validator")
        print("  \___________________________________________________________________")
        ip_addr_list, username_list, password_list = device_info_validator(device_info_list)
    # Stop process by keyboard (e.g. CTRL+C)
    except KeyboardInterrupt:
        # Execute program_cancellation
        program_cancellation()

    # Execute ip_addr_validator
    try:
        print("\n \  IP address validator")
        print("  \___________________________________________________________________")
        # Verify the first try of validation with a flag. Return none if correct or return a new list if corrected
        ip_addr_list_fix = ip_addr_validator("PASS", ip_addr_list)
        # Execute ip_addr_list_validator
        ip_addr_list = ip_addr_list_validator(ip_addr_list_fix, ip_addr_list)
    # Stop process by keyboard (e.g. CTRL+C)
    except KeyboardInterrupt:
        # Execute program_cancellation
        program_cancellation()

    # Execute connection_validator
    try:
        print("\n \  Connection validator powered by", powered_by("pythonping"))
        print("  \___________________________________________________________________")
        # Verify the first try of validation with a flag. Return none if correct or return a new list if corrected
        ip_addr_list_fix = connection_validator("PASS", ip_addr_list)
        # Execute ip_addr_list_validator
        ip_addr_list = ip_addr_list_validator(ip_addr_list_fix, ip_addr_list)
    # Stop process by keyboard (e.g. CTRL+C)
    except KeyboardInterrupt:
        # Execute program_cancellation
        program_cancellation()

    # Execute user confirmation to execute connection_futures
    try:
        print("\n \  SSH remote configuration powered by", powered_by("netmiko"))
        print("  \___________________________________________________________________\n")
        print("Confirmation to proceed:")
        # Execute enter_key_confirmation
        enter_key_confirmation()
        connection_futures(device, command, delay, ip_addr_list, username_list, password_list, command_list)
    # Stop process by keyboard (e.g. CTRL+C)
    except KeyboardInterrupt:
        # Execute program_cancellation
        print("")
        program_cancellation()

    # Execute exit_and_cleanup
    exit_and_cleanup()


import re

# Import function from module
from .file_validator import device_info_list_validator

def is_complete(device_info_elements):
    # Device information must have 3 elements
    result = len(device_info_elements) == 3
    # Pass the value
    return result

def is_empty(device_info_elements):
    if is_complete(device_info_elements):
        regex_results = []
        # No-empty element patterns
        regex_results.append(not re.match(r"^.{0}$", device_info_elements[0]))
        regex_results.append(not re.match(r"^.{0}$", device_info_elements[1]))
        regex_results.append(not re.match(r"^.{0}$", device_info_elements[2]))
        # When all patterns match, all Results should be True
        if all(regex_results):
            result = True
        else:
            result = False
        # Pass the value
        return result
    # Invalid device information if the elements are not 3
    else:
        # Pass the value
        return False

def device_info_validator(device_info_list):
    # Progress dot
    print("\nValidating device(s) information. Please wait", end = "", flush = True)
    # An empty device information list is not accepted
    if len(device_info_list) == 0:
        print("\n\nFAIL: The device information list is empty!")
        # Repeat execute device_info_validator and then pass the value
        return device_info_validator(device_info_list_validator())
    else:
        # Initial variables
        ip_addr_list = []
        username_list = []
        password_list = []
        invalid_formats = []
        results = []
        # Loop for every device information in the list
        for device_info in device_info_list:
            # Strip newline at the end of the device information then split the device information into IP address, username, and password format
            device_info_elements = device_info.rstrip("\n").split("::")
            # Device information must have an IP address, a username, and a password
            if is_empty(device_info_elements):
                results.append("PASS")
                # Split elements of the device information and store them in their new group list
                ip_addr_list.append(device_info_elements[0])
                username_list.append(device_info_elements[1])
                password_list.append(device_info_elements[2])
                # Progress dot
                print(".", end = "", flush = True)
                continue
            # Device information that is not complete or contains an empty IP address or/and username or/and password is invalid
            else:
                results.append("FAIL")
                invalid_formats.append(device_info.rstrip("\n"))
                # Progress dot
                print(".", end = "", flush = True)
                continue
        # Unacceptable results
        if "FAIL" in results:
            # Initial variables
            obscured_formats = []
            # Loop for every device information in the list
            for device_info in invalid_formats:
                # Obscure device information
                obscured_info = re.sub(r"([^::])", "*", device_info, 0, re.MULTILINE)
                # Store obscured device information in the new list
                obscured_formats.append(obscured_info)
            print("\n\nFAIL: Invalid device(s) information format: " + str(obscured_formats)[1:-1])
            print("FAIL: Please correct the device information list and try again!")
            # Repeat execute device_info_validator and then pass the value
            return device_info_validator(device_info_list_validator())
        # Acceptable results
        else:
            print("\n\nPASS: The device information list is valid")
            # Pass these values
            return ip_addr_list, username_list, password_list


import os

def command_type_validator(flag, input_data):
    # Command type send_config_set and Device information list are treatable in the same manner
    if (flag == "DEVICE") or (flag == "CONFIG_2"):
        # Pass the value
        return input_data
    # Command type send_command only accepting one command
    elif (flag == "CONFIG_1") and (len(input_data) == 1):
        # Pass the value
        return input_data[0].rstrip("\n")
    # Command type send_command not accepting multiple commands
    else:
        print("FAIL: There are " + str(len(input_data)) + " commands found in the configuration file")
        print("FAIL: Please use only one command for command type \'send_command\' and try again!")
        # Repeat execute configuration_file_validator with the same flag and then pass the value
        return configuration_file_validator(flag)

def file_validator(flag, input_file):
    # File exist is accepted
    if os.path.isfile(input_file) == True:
        print("PASS: \'" + input_file + "\' found")
        # Open the file with read permission
        with open(input_file, "r") as file:
            # Read the file from the beginning of the file
            file.seek(0)
            # Verify the configuration file's content according to the command type flag and then pass the value
            return command_type_validator(flag, file.readlines())
    # File not exist is not accepted
    else:
        print("FAIL: \'{}\' not found!".format(input_file))
        if flag == "DEVICE":
            # Repeat execute device_info_list_validator and then pass the value
            return device_info_list_validator()
        elif (flag == "CONFIG_1") or (flag == "CONFIG_2"):
            # Repeat execute configuration_file_validator with the same flag and then pass the value
            return configuration_file_validator(flag)

def device_info_list_validator():
    # Prompt user for
    device_info_list = input("\nEnter the path and the name of the device information list file: ")
    # Verify the file's existence with a flag and then pass the value
    return file_validator("DEVICE", device_info_list)

def configuration_file_validator(command):
    # Flag the configuration file based on the command type
    if command == "send_command":
        flag = "CONFIG_1"
    elif command == "send_config_set":
        flag = "CONFIG_2"
    # When configuration_file_validator executed by command_type_validator; Use the same flag
    else:
        flag = command
    # Prompt user for
    configuration_file = input("\nEnter the path and the name of the configuration file: ")
    # Verify the file's existence with a flag and then pass the value
    return file_validator(flag, configuration_file)


import re

# Import function from module
from .program_supplementals import input_option, powered_by

def supported_device_types():
    # For supported device types information, please visit 'https://github.com/ktbyers/netmiko/blob/master/netmiko/ssh_dispatcher.py' in CLASS_MAPPER_BASE dictionary
    supported_device_types = ["a10", "accedian", "adtran_os", "alcatel_aos", "alcatel_sros", "allied_telesis_awplus", "apresia_aeos", "arista_eos", "aruba_os", "aruba_osswitch", "aruba_procurve", "avaya_ers", "avaya_vsp", "broadcom_icos", "brocade_fos", "brocade_fastiron", "brocade_netiron", "brocade_nos", "brocade_vdx", "brocade_vyos", "checkpoint_gaia", "calix_b6", "cdot_cros", "centec_os", "ciena_saos", "cisco_asa", "cisco_ftd", "cisco_ios", "cisco_nxos", "cisco_s300", "cisco_tp", "cisco_wlc", "cisco_xe", "cisco_xr", "cloudgenix_ion", "coriant", "dell_dnos9", "dell_force10", "dell_os6", "dell_os9", "dell_os10", "dell_powerconnect", "dell_isilon", "dlink_ds", "endace", "eltex", "eltex_esr", "enterasys", "ericsson_ipos", "extreme", "extreme_ers", "extreme_exos", "extreme_netiron", "extreme_nos", "extreme_slx", "extreme_vdx", "extreme_vsp", "extreme_wing", "f5_ltm", "f5_tmsh", "f5_linux", "flexvnf", "fortinet", "generic", "generic_termserver", "hp_comware", "hp_procurve", "huawei", "huawei_smartax", "huawei_olt", "huawei_vrpv8", "ipinfusion_ocnos", "juniper", "juniper_junos", "juniper_screenos", "keymile", "keymile_nos", "linux", "mikrotik_routeros", "mikrotik_switchos", "mellanox", "mellanox_mlnxos", "mrv_lx", "mrv_optiswitch", "netapp_cdot", "netgear_prosafe", "netscaler", "nokia_sros", "oneaccess_oneos", "ovs_linux", "paloalto_panos", "pluribus", "quanta_mesh", "rad_etx", "raisecom_roap", "ruckus_fastiron", "ruijie_os", "sixwind_os", "sophos_sfos", "supermicro_smis", "tplink_jetstream", "ubiquiti_edge", "ubiquiti_edgerouter", "ubiquiti_edgeswitch", "ubiquiti_unifiswitch", "vyatta_vyos", "vyos", "watchguard_fireware", "zte_zxros", "yamaha"]
    # Pass the value
    return supported_device_types

def command_types():
    # For more information about these command types, please visit 'https://ktbyers.github.io/netmiko/docs/netmiko/index.html#netmiko.BaseConnection.send_command' and 'https://ktbyers.github.io/netmiko/docs/netmiko/index.html#netmiko.BaseConnection.send_config_set'
    command_types = ["send_command", "send_config_set"]
    # Pass the value
    return command_types

def device_type_check():
    # Prompt user for
    device_type = input("Enter the name of the device type: ")
    # Device type match is accepted
    if device_type in supported_device_types():
        print("PASS: \'{}\' is supported".format(device_type), "\n")
        # Pass the value
        return device_type
    # Device type not match is not accepted
    else:
        print("FAIL: \'{}\' is not supported!".format(device_type), "\n")
        # Repeat execute configuration_type_validator and then pass the value
        return device_type_check()

def command_type_check():
    # Prompt user for
    command_type = input("Enter the name of the command type: ")
    # Command type match is accepted
    if command_type in command_types():
        print("PASS: \'{}\' is valid".format(command_type), "\n")
        # Pass the value
        return command_type
    # Command type not match is not accepted
    else:
        print("FAIL: \'{}\' is invalid!".format(command_type), "\n")
        # Repeat execute configuration_type_validator and then pass the value
        return command_type_check()

def global_delay_factor():
    # Prompt user for
    delay = input("Enter the delay value [0.1–100] or press \'Enter\' for default value [1.2]: ")
    # Zero value is not accepted
    if bool(re.match(r"(^(0|00)[.](0|00)$|^00$|^0$)", delay)):
        print("FAIL: \'{}\' is invalid!".format(delay), "\n")
        # Repeat execute global_delay_factor and then pass the value
        return global_delay_factor()
    # Value 0.1–100 and empty string 'enter key' is accepted
    elif bool(re.match(r"((^([1-9]|[0-9][0-9])[.]([0-9]|[0-9][0-9])$)|(^100$)|(^[0-9][0-9]$)|(^[1-9]$))", delay)) or bool(delay == ""):
        # Default value
        if (not delay) or (delay == "1.2") or (delay == "1.20"):
            print("PASS: global_delay_factor: default\n")
            # Pass the default value
            return float(1.2)
        # The default value of netmiko
        elif (delay == "1") or (delay == "1.0") or (delay == "1.00"):
            print("PASS: global_delay_factor: netmiko's default\n")
            # Pass the netmiko's value
            return float(delay)
        # User value
        else:
            print("PASS: global_delay_factor: \'" + str(float(delay)) + "\'\n")
            # Pass the user value
            return float(delay)
    # A value greater than 100, trailing and leading zero more than 2 is not accepted
    else:
        print("FAIL: \'{}\' is invalid!".format(delay), "\n")
        # Repeat execute global_delay_factor and then pass the value
        return global_delay_factor()

def configuration_type_validator():
    print(" \  Configuration type validator")
    print("  \___________________________________________________________________\n")
    print("[A] Enter the device type and the command type")
    print("[B] Show the list of supported device type")
    print("[C] Show the list of supported command type")
    print("\nPlease choose one of the options above or Press \'CTRL+C\' to abort the program: ", end="", flush=True)
    # Capture any inputs and suppress the output
    key = input_option()
    # Compare the chosen option with the available options
    if (key == "a") or (key == "A"):
        print("")
        # Verify the support of the device type and validity of the command type
        device_type = device_type_check()
        command_type = command_type_check()
        # If the user has chosen send_command, use the netmiko's global_delay_factor default value
        if bool(command_type == "send_command"):
            return device_type, command_type, float(1.0)
        # If the user has chosen send_config_set, execute global_delay_factor
        elif bool(command_type == "send_config_set"):
            return device_type, command_type, global_delay_factor()
    elif (key == "b") or (key == "B"):
        print("\n \  Supported device types are powered by", powered_by("netmiko"))
        print("  \___________________________________________________________________\n")
        # Loop for every device type in the supported device types list
        for device_type in supported_device_types():
            print(">", device_type)
        print("")
        # Repeat execute configuration_type_validator and then pass the value
        return configuration_type_validator()
    elif (key == "c") or (key == "C"):
        print("\n \  Command types are powered by", powered_by("netmiko"))
        print("  \___________________________________________________________________\n")
        print(">", command_types()[0] + "\n  TL;DR      : Send one command and wait until the information is entirely received.\n  Explanation: Use to execute a command on the SSH channel to show information.\n               This method will keep waiting to receive data until the network device prompt is detected.")
        print(">", command_types()[1] + "\n  TL;DR      : Send one or a set of commands in configuration mode.\n  Explanation: Use to send configuration commands down the SSH channel and will be executed one after the other.\n               Automatically exits/enters configuration mode.")
        print("  >> global_delay_factor\n     TL;DR      : Set a multiplication value for additional delays in netmiko's command execution.\n     Explanation: Multiplication factor affecting Netmiko delays with the default value is set on \'1.0\'.\n                  The higher value will result in the program being slow in the process hence too long to complete.\n                  This program wisely set the default value on \'1.2\'.\n")
        # Repeat execute configuration_type_validator and then pass the value
        return configuration_type_validator()
    else:
        print("\nFAIL: Please choose \'A\' or \'B\' or \'C\'!\n")
        # Repeat execute configuration_type_validator and then pass the value
        return configuration_type_validator()


from setuptools import setup

setup(
    name = "Malas",
    version = "1.0.210926.0001",
    description = "Malas (Multi-vendor automation leverage at slouching)",
    url = "https://github.com/destrianto/malas",
    download_url = "https://github.com/destrianto/malas/releases/download/v1.0/malas_1.0.210926.0001.zip",
    author = "Ade Destrianto",
    author_email = "ade.destrianto@mail.ru",
    license = "MIT",
    classifiers = [
        "Environment :: Console",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.9",
        "Topic :: System :: Networking"
    ],
    keywords = [
        "multi-vendor",
        "network",
        "automation"
    ],
    packages = [
        "malas"
    ],
    install_requires = [
        "netmiko>=3.4.0",
        "pythonping>=1.1.0"
    ],
    python_requires = ">=3.9"
)


import re

# Import function from module
from .file_validator import device_info_list_validator
from .device_info_validator import device_info_validator

def is_octet(ip_octets):
    # IP address must have 4 octets
    result = len(ip_octets) == 4
    # Pass the value
    return result

def is_empty(ip_octets):
    if is_octet(ip_octets):
        regex_results = []
        # No-empty octet patterns
        regex_results.append(not re.match(r"^.{0}$", ip_octets[0]))
        regex_results.append(not re.match(r"^.{0}$", ip_octets[1]))
        regex_results.append(not re.match(r"^.{0}$", ip_octets[2]))
        regex_results.append(not re.match(r"^.{0}$", ip_octets[3]))
        # When all patterns match, all Results should be True
        if all(regex_results):
            result = True
        else:
            result = False
        # Pass the value
        return result
    # Invalid IP address if the octets are not 4
    else:
        # Pass the value
        return False

def is_digit(ip_octets):
    if is_octet(ip_octets):
        regex_results = []
        # Octet patterns contain no other than digit
        regex_results.append(not re.match(r"(?!^\d+$)^.+$", ip_octets[0]))
        regex_results.append(not re.match(r"(?!^\d+$)^.+$", ip_octets[1]))
        regex_results.append(not re.match(r"(?!^\d+$)^.+$", ip_octets[2]))
        regex_results.append(not re.match(r"(?!^\d+$)^.+$", ip_octets[3]))
        # When all patterns match, all Results should be True
        if all(regex_results):
            result = True
        else:
            result = False
        # Pass the value
        return result
    # Invalid IP address if the octets are not 4
    else:
        # Pass the value
        return False

def is_unspecified_address(ip_octets):
    regex_results = []
    # Unspecified address patterns
    regex_results.append(re.match(r"0", ip_octets[0]))
    regex_results.append(re.match(r"0", ip_octets[1]))
    regex_results.append(re.match(r"0", ip_octets[2]))
    regex_results.append(re.match(r"0", ip_octets[3]))
    # When all patterns match, all Results should be True
    if all(regex_results):
        result = True
    else:
        result = False
    # Pass the value
    return result

def is_linklocal_address(ip_octets):
    regex_results = []
    # Link-local address patterns
    regex_results.append(re.match(r"169", ip_octets[0]))
    regex_results.append(re.match(r"254", ip_octets[1]))
    regex_results.append(re.match(r"([0-9]|[1-9][0-9]|[1-2][0-9][0-9])", ip_octets[2]))
    regex_results.append(re.match(r"([0-9]|[1-9][0-9]|[1-2][0-9][0-9])", ip_octets[3]))
    # When all patterns match, all Results should be True
    if all(regex_results):
        result = True
    else:
        result = False
    # Pass the value
    return result

def is_localhost_address(ip_octets):
    regex_results = []
    # Localhost address patterns
    regex_results.append(re.match(r"127", ip_octets[0]))
    regex_results.append(re.match(r"([0-9]|[1-9][0-9]|[1-2][0-9][0-9])", ip_octets[1]))
    regex_results.append(re.match(r"([0-9]|[1-9][0-9]|[1-2][0-9][0-9])", ip_octets[2]))
    regex_results.append(re.match(r"([0-9]|[1-9][0-9]|[1-2][0-9][0-9])", ip_octets[3]))
    # When all patterns match, all Results should be True
    if all(regex_results):
        result = True
    else:
        result = False
    # Pass the value
    return result

def is_multicast_address(ip_octets):
    regex_results = []
    # Multicast address patterns
    regex_results.append(re.match(r"(22[4-9]|23[0-5]|239)", ip_octets[0]))
    regex_results.append(re.match(r"([0-9]|[1-9][0-9]|[1-2][0-9][0-9])", ip_octets[1]))
    regex_results.append(re.match(r"([0-9]|[1-9][0-9]|[1-2][0-9][0-9])", ip_octets[2]))
    regex_results.append(re.match(r"([0-9]|[1-9][0-9]|[1-2][0-9][0-9])", ip_octets[3]))
    # When all patterns match, all Results should be True
    if all(regex_results):
        result = True
    else:
        result = False
    # Pass the value
    return result

def is_unassigned_address(ip_octets):
    regex_results = []
    # Unassigned address patterns
    regex_results.append(re.match(r"(24[0-9]|25[0-5])", ip_octets[0]))
    regex_results.append(re.match(r"([0-9]|[1-9][0-9]|[1-2][0-9][0-9])", ip_octets[1]))
    regex_results.append(re.match(r"([0-9]|[1-9][0-9]|[1-2][0-9][0-9])", ip_octets[2]))
    regex_results.append(re.match(r"([0-9]|[1-9][0-9]|[1-2][0-9][0-9])", ip_octets[3]))
    # When all patterns match, all Results should be True
    if all(regex_results):
        result = True
    else:
        result = False
    # Pass the value
    return result

def ip_addr_validator(flag, ip_addr_list):
    # Progress dot
    print("\nValidating IP address(es). Please wait", end = "", flush = True)
    # Initial variables
    valid_ips = []
    invalid_ips = []
    invalid_formats = []
    results = []
    # Loop for every IP address in the list
    for ip in ip_addr_list:
        # Strip newline at the end of IP address then split IP address to octets
        ip_octets = ip.rstrip("\n").split(".")
        # IP address must have 4 octets, an octet cannot be empty, and octets must be a digit
        if all([is_empty(ip_octets), is_digit(ip_octets)]):
            # Network address other than Private address and Public address is not allowed
            if any([is_unspecified_address(ip_octets), is_linklocal_address(ip_octets), is_localhost_address(ip_octets), is_multicast_address(ip_octets), is_unassigned_address(ip_octets)]):
                results.append("FAIL")
                invalid_ips.append(ip)
                # Progress dot
                print(".", end = "", flush = True)
                continue
            # Private address or Public address is allowed
            else:
                results.append("PASS")
                valid_ips.append(ip)
                # Progress dot
                print(".", end = "", flush = True)
                continue
        # IP address must have 4 octets, an octet cannot be empty, and octets must be a digit
        else:
            results.append("FAIL")
            invalid_formats.append(ip)
            # Progress dot
            print(".", end = "", flush = True)
            continue
    # Unaccepted IP address list results
    if "FAIL" in results:
        print("\n\nPASS: Valid IP address(es): " + str(valid_ips)[1:-1])
        print("FAIL: Invalid IP address(es): " + str(invalid_ips)[1:-1])
        print("FAIL: Invalid IP address(es) format: " + str(invalid_formats)[1:-1])
        print("FAIL: Please correct the IP address list and try again!")
        # Repeat execute device_info_list_validator then device_info_validator and then ip_addr_validator with a flag
        return ip_addr_validator("FAIL", device_info_validator(device_info_list_validator())[0])
    # Accepted IP address list results
    else:
        # IP address list with correction flag
        if flag == "FAIL":
            print("\n\nPASS: The IP address list is valid")
            # Pass the value
            return ip_addr_list
        # IP address list with no correction flag
        elif flag == "PASS":
            print("\n\nPASS: The IP address list is valid")

def ip_addr_list_validator(ip_addr_list_fix, ip_addr_list):
    # If a new list exists, replace the existing ip_addr_list with the corrected one
    if bool(ip_addr_list_fix) == True:
        ip_addr_list = ip_addr_list_fix
        # Pass the value
        return ip_addr_list
    # If there's no new list exists, pass the existing list
    else:
        # Pass the value
        return ip_addr_list


import os
import sys
import time
import random
import string
import datetime
import concurrent.futures

# Import function from module
from .program_supplementals import exit_and_cleanup, enter_key_only, exception_translator

# Import function from 3rd party module
from netmiko import ConnectHandler

def file_output(ssh_results, ssh_success, ssh_failed):
    # Get the current path of the running Python file
    current_path = os.path.dirname(os.path.realpath(__file__))
    # Prompt user for
    target_path = input("\nEnter the target path or leave it blank to set the default path [" + current_path + "]: ")
    # If target_path is blank, fill it with a default directory name
    if bool(target_path == ""):
        target_path = "Malas_SSH_outputs"
    try:
        # Create a new directory if not exists yet on the target path to contains all SSH output file(s)
        if bool(os.path.exists(target_path)) == False:
            os.makedirs(target_path)
        # Loop for every result in the list
        for ssh_result in ssh_results:
            # Give a unique key for the output file
            unique_key = "".join(random.choice(string.ascii_uppercase + string.digits) for _ in range(6))
            # Get the current date and time
            present = datetime.datetime.now().strftime("_on_%Y-%m-%d_at_%H.%M")
            # Merge target path with the file name and its extension
            complete_path = os.path.join(target_path, ssh_result[0] + present + "_[" + unique_key + "].txt")
            # Open the file with write permission
            with open(complete_path, "w") as file:
                # Write the SSH outputs to the file
                file.write("%s" % ssh_result[1])
        # SSH attempt results
        print("\nSSH remote configuration success: " + str(ssh_success) + " host(s)")
        print("SSH remote configuration failed: " + str(ssh_failed) + " host(s)")
        # target_path is the default directory name
        if bool(target_path == "Malas_SSH_outputs"):
            print("\nPASS: The SSH output file(s) are stored in the path \'" + current_path + "\' inside the directory \'" + target_path + "\' successfully")
        # target_path is user-defined
        else:
            print("\nPASS: The SSH output file(s) are stored in the path \'" + target_path + "\' successfully")
        print("EXIT: Please review the SSH output file(s) to confirm the configured configuration, thank you!")
    except:
        # Execute exception_translator
        exception_explained = exception_translator()
        # Print the raised exception error messages values
        print("\nFAIL: " + exception_explained[0] + ":\n" + exception_explained[1])
        # Repeat execute file_output and then pass these values
        file_output(ssh_results, ssh_success, ssh_failed)

def thread_processor(threads):
    # Initial variables
    ssh_results = []
    ssh_success = 0
    ssh_failed = 0
    # Loop for every result from ssh-threading process
    for thread in threads:
        # Store the thread results values
        ssh_result = thread.result()
        # Failed SSH attempts contain 2 values in tuple formats
        if isinstance(ssh_result[1], tuple):
            # Merge raised exception error name and explanation
            result_concatenated = "FAIL: " + ssh_result[1][0] + "\n\n" + ssh_result[1][1]
            # Store the raised exception error messages values in the same index
            ssh_results.append((ssh_result[0], result_concatenated))
            # Increment of failed SSH attempts
            ssh_failed += 1
        else:
            # Store the raised exception error messages values
            ssh_results.append(ssh_result)
            # Increment of success SSH attempts
            ssh_success += 1
    try:
        # Execute user confirmation to create output file(s)
        print("\nPress \'Enter\' to create the SSH output file(s) or \'CTRL+C\' to end the program", end = "", flush = True)
        # Expect the user to press Enter key
        enter_key_only()
        # Execute file_output
        file_output(ssh_results, ssh_success, ssh_failed)
    # Stop process by keyboard (e.g. CTRL+C)
    except KeyboardInterrupt:
        # SSH attempt results
        print("\n\nSSH remote configuration success: " + str(ssh_success) + " host(s)")
        print("SSH remote configuration failed: " + str(ssh_failed) + " host(s)")
        print("\nEXIT: Please review the SSH outputs to confirm the configured configuration, thank you!")
        # Execute exit_and_cleanup
        exit_and_cleanup()

def output_processor(output, command, stopwatch):
    # Remote configuration stopwatch end
    ssh_processed = "\'%.2f\'" % (time.time() - stopwatch) + " secs"
    # Process the output according to its command type
    if command == "send_command":
        # No output process
        final_output = output
    elif command == "send_config_set":
        # Split output into a list
        disintegrate_output = output.split("\n")
        # Remove the unnecessary lines
        final_output = "\n".join(disintegrate_output[1:-1])
    # Pass these values
    return final_output, ssh_processed

def connection_ssh(dev, cmd, gdf, ip, usr, pwd, cfg):
    # Strip newline at the end of device type, command type, IP address, username, and password
    device = dev.rstrip("\n")
    command = cmd.rstrip("\n")
    ip_addr = ip.rstrip("\n")
    username = usr.rstrip("\n")
    password = pwd.rstrip("\n")
    try:
        # Remote configuration stopwatch start
        stopwatch = time.time()
        # Define the device type, the credential information, and the delay value to log in to the remote host
        session = {
            "device_type": device,
            "host": ip_addr,
            "username": username,
            "password": password,
            "global_delay_factor": gdf
        }
        # SSH to the remote host
        remote = ConnectHandler(**session)
        # Execute every command in the configuration file according to its command type
        if command == "send_command":
            output = remote.send_command(cfg)
            # Execute output_processor and retrive values
            final_output, ssh_processed = output_processor(output, command, stopwatch)
        elif command == "send_config_set":
            output = remote.send_config_set(cfg)
            # Execute output_processor and retrive values
            final_output, ssh_processed = output_processor(output, command, stopwatch)
        # Output's bracket and print the output
        print("\n\n \  Remote host \'" + ip_addr + "\' processed for " + ssh_processed + "\n  \___________________________________________________________________\n\n" + final_output, end="")
        # Pass values to threading result
        return ip_addr, final_output
    except:
        # Execute exception_translator
        exception_explained = exception_translator()
        # Output's bracket and print the output
        print("\n\n \  Remote host \'" + ip_addr + "\' failed to configure\n  \___________________________________________________________________\n\nFAIL: " + exception_explained[0] + "\n\n" + exception_explained[1], end = "")
        # Pass values to threading result
        return ip_addr, exception_explained

def connection_futures(device, command, delay, ip_addr_list, username_list, password_list, command_list):
    # Execute connection_ssh. Progress dot with threading capability
    print("\nConcurrently configuring per", min(32, os.cpu_count() + 4), "hosts. Please wait", end = "", flush = True)
    # SSH-threading stopwatch start
    threading_start = time.time()
    # Suppress raised exception error messages outputs
    sys.stderr = os.devnull
    # SSH-threading process
    with concurrent.futures.ThreadPoolExecutor() as executor:
        # Initial variables
        threads = []
        ssh_attempts = 0
        # Loop for every IP address, username, and password in the list
        for ip_addr, username, password in zip(ip_addr_list, username_list, password_list):
            # Increment of SSH attempts
            ssh_attempts += 1
            # Execute configuration over SSH for every IP address, username, and password in the list concurrently
            threads.append(executor.submit(connection_ssh, dev = device, cmd = command, gdf = delay, ip = ip_addr, usr = username, pwd = password, cfg = command_list))
            # Progress dot
            print(".", end = "", flush = True)
    # Unsuppress raised exception error messages outputs
    sys.stderr = sys.__stderr__
    print("\n\n \  Completed")
    print("  \___________________________________________________________________\n")
    # SSH attempt results and ping-threading stopwatch end
    print("SSH-threading for " + str(ssh_attempts) + " host(s) processed for:", "%.2f" % (time.time() - threading_start), "secs")
    # Execute thread_processor
    thread_processor(threads)


import os
import sys
import shutil
import getpass
import subprocess
import pkg_resources

def exit_and_cleanup():
    target = os.path.join("malas", "__pycache__")
    # Delete pycache folder and its content
    shutil.rmtree(target)
    # Exit program
    sys.exit()

def enter_key_only():
    # Expect the user to press Enter key and suppress the output
    getpass.getpass("")

def enter_key_confirmation():
    print("Press \'Enter\' to continue or \'CTRL+C\' to abort the program", end="", flush=True)
    # Expect the user to press Enter key and suppress the output
    getpass.getpass("")

def input_option():
    # Capture any inputs
    return str(input(""))

def exception_translator():
    # Get the raised exception error messages values
    exc_type, exc_value, _ = sys.exc_info()
    # Store the raised exception error messages values
    exception_name = exc_type.__name__
    exception_explanation = str(exc_value)
    # Output for blank raised exception error explanation
    if len(exception_explanation) == 0:
        exception_explanation = "There's no explanation provided for this exception."
    # Pass these values
    return exception_name, exception_explanation

def module_verifier(module):
    try:
        # Get installed module's name
        module_name = pkg_resources.get_distribution(module).key
        # Get installed module's version
        module_version = pkg_resources.get_distribution(module).version
        # Pass these values
        return module_name, module_version
    except:
        # Pass these values when the module is not installed
        return False, False

def install_requirements(module):
    print("Installing required module: ", end="", flush=True)
    try:
        # Install the module and suppress outputs
        subprocess.check_call([sys.executable, "-m", "pip", "install", module], stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT)
        print("Done, " + module + " " + module_verifier(module)[1], flush=True)
        return True
    # Module installations failed due to an Internet problem
    except:
        print("Failed (" + module + ")", flush=True)
        return False

def program_requirements():
    # Required modules
    prerequisites = ["netmiko", "pythonping", "bcrypt", "cffi", "cryptography", "future", "ntc-templates", "paramiko", "pycparser", "pynacl", "pyserial", "scp", "setuptools", "six", "tenacity", "textfsm"]
    # Initial variables
    module_results = []
    module_installs = []
    # Loop for every required module in the list
    for module in prerequisites:
        # Verify if modules are installed and store results in a list
        module_results.append([module, bool(module_verifier(module)[0])])
    # Loop for every module check results. If the module is not installed, store the module name in a list
    if all([module_installs.append(result[0]) if result[1] == False else True for result in module_results]):
        # All required modules are installed
        pass
    # Install the required modules
    else:
        print("\n \  Self-diagnostics and Self-recovery")
        print("  \___________________________________________________________________\n")
        # Initial variables
        install_results = []
        # Loop for every module in the list
        for module in module_installs:
            # Execute install_requirements to install the required modules and store results in a list
            install_results.append(install_requirements(module))
        # Loop for every module installation result
        if all([True if result == True else False for result in install_results]):
            print("\nDiagnostics and recovery are completed")
        # Module installations failed due to an Internet problem
        else:
            print("\nPlease check the Internet connection and try again!")
            print("Alternatively, please perform manual module installation!")
            # Execute exit_and_cleanup
            exit_and_cleanup()

def powered_by(module):
    # Execute module_verifier to get the installed module's name and version
    module_name, module_version = module_verifier(module)
    # Pass the value
    return (module_name + " " + module_version)

def program_cancellation():
    print("\nEXIT: I\'ll see you again :)")
    # Execute exit_and_cleanup
    exit_and_cleanup()